import datetime as dt

from marshmallow import Schema, fields, post_load, pre_load

from boinc_client.models.helpers import (
    create_indexes,
    flatten_data,
    normalise_none_to_list,
    remove_key,
)


def _epoch_to_date(epoch_days: str) -> str:
    epoch = dt.date(1970, 1, 1)
    delta = dt.timedelta(days=int(epoch_days))
    return (epoch + delta).strftime("%Y-%m-%d")


class Transfer(Schema):
    down = fields.Float()
    up = fields.Float()
    when = fields.Int()

    @post_load
    def _a_remove_when(self, data, **kwargs):
        return remove_key(data, "when")


class DailyTransfers(Schema):
    network_transfers = fields.Dict(
        fields.Str(), fields.Nested(Transfer()), data_key="daily_xfers"
    )

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "daily_xfers")

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "daily_xfers", "dx")

    @pre_load
    def _c_create_dict_keys(self, data, **kwargs):
        data["daily_xfers"] = {
            _epoch_to_date(dx["when"]): dx for dx in data["daily_xfers"]
        }
        return data



from marshmallow import Schema, fields


class Update(Schema):
    download_url = fields.Url()
    newer_version = fields.Str(allow_none=True)


class ClientUpdate(Schema):
    update = fields.Nested(Update())



import xmltodict

from boinc_client.clients.rpc_client import RpcClient
from boinc_client.models.generic_response import GenericResponse

VALID_MODES = ["always", "never", "auto", "restore"]


def set_cpu_run_mode(client: RpcClient, run_mode: str, duration: int = 0):
    if run_mode in VALID_MODES:
        req_string = f"<set_run_mode><{run_mode}/>"
        if duration > 0:
            req_string += f"<duration>{duration}</duration>"
        req_string += "</set_run_mode>"
        rpc_resp = client.make_request(req_string)
        rpc_json = xmltodict.parse(rpc_resp)
        return GenericResponse().load(rpc_json)
    return {
        "error": "Invalid CPU mode given. Select from: always, never, auto or restore"
    }


def set_gpu_run_mode(client: RpcClient, run_mode: str, duration: int = 0):
    if run_mode in VALID_MODES:
        req_string = f"<set_gpu_mode><{run_mode}/>"
        if duration > 0:
            req_string += f"<duration>{duration}</duration>"
        req_string += "</set_gpu_mode>"
        rpc_resp = client.make_request(req_string)
        rpc_json = xmltodict.parse(rpc_resp)
        return GenericResponse().load(rpc_json)
    return {
        "error": "Invalid GPU mode given. Select from: always, never, auto or restore"
    }


def set_network_mode(client: RpcClient, run_mode: str, duration: int = 0):
    if run_mode in VALID_MODES:
        req_string = f"<set_network_mode><{run_mode}/>"
        if duration > 0:
            req_string += f"<duration>{duration}</duration>"
        req_string += "</set_network_mode>"
        rpc_resp = client.make_request(req_string)
        rpc_json = xmltodict.parse(rpc_resp)
        return GenericResponse().load(rpc_json)
    return {
        "error": "Invalid network mode given. Select from: always, never, auto or restore"
    }



from marshmallow import Schema, fields, post_load, pre_load

from boinc_client.models.helpers import normalise_none_to_list


class CoProc(Schema):
    name: fields.Str()


class Host(Schema):
    coprocs = fields.Nested(CoProc(many=True))
    d_free = fields.Float()
    d_total = fields.Float()
    domain_name = fields.Str()
    host_cpid = fields.Str()
    ip_addr = fields.Str()
    m_cache = fields.Float()
    m_nbytes = fields.Float()
    m_swap = fields.Float()
    n_usable_coprocs = fields.Int()
    os_name = fields.Str()
    os_version = fields.Str()
    p_calculated = fields.Float()
    p_features = fields.Str()
    p_fpops = fields.Float()
    p_iops = fields.Float()
    p_membw = fields.Float()
    p_model = fields.Str(allow_none=True, load_default="")
    p_ncpus = fields.Int()
    p_vendor = fields.Str(allow_none=True, load_default="")
    p_vm_extensions_disabled = fields.Bool()
    timezone = fields.Int()
    wsl_available = fields.Bool()

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "coprocs")

    @post_load
    def _a_replace_none_string(self, data, **kwargs):
        for field in data:
            data[field] = data[field] if data[field] is not None else ""
        return data


class HostInfo(Schema):
    host_info = fields.Nested(Host())



from marshmallow import Schema, fields, pre_load

from boinc_client.models.helpers import flatten_data, normalise_none_to_list


class ProjectPlatform(Schema):
    name = fields.Str()


class Project(Schema):
    description = fields.Str()
    general_area = fields.Str()
    home = fields.Str()
    id = fields.Int()
    image = fields.Str()
    keywords = fields.Str()
    name = fields.Str()
    platforms = fields.Nested(ProjectPlatform(many=True))
    specific_area = fields.Str()
    summary = fields.Str()
    url = fields.Url()
    web_url = fields.Url()

    @pre_load
    def _convert_platform_dict_to_list(self, data, **kwargs):
        data["platforms"] = (
            [{"name": d} for d in data["platforms"]["name"]]
            if type(data["platforms"]["name"]) == list
            else [data["platforms"]]
        )
        return data


class Projects(Schema):
    projects = fields.Nested(Project(many=True))

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "projects")

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "projects", "project")



import xmltodict

from boinc_client.clients.rpc_client import RpcClient
from boinc_client.models.message_count import MessageCount
from boinc_client.models.messages import Messages
from boinc_client.models.public_notice import Notices


def messages(client: RpcClient, start: int = 0) -> dict:
    """Show messages with sequence numbers beyond the given seqno."""
    rpc_resp = client.make_request(
        f"<get_messages><seqno>{start}</seqno></get_messages>"
    )
    rpc_json = xmltodict.parse(rpc_resp, force_list="msg")
    return Messages().load(rpc_json)


def message_count(client: RpcClient) -> dict:
    """Show the largest message seqno."""
    rpc_resp = client.make_request("<get_message_count/>")
    rpc_json = xmltodict.parse(rpc_resp)
    return MessageCount().load(rpc_json)


def public_notices(client: RpcClient, start: int = 1) -> dict:
    """Returns only non-private notices, doesn't require authentication."""
    rpc_resp = client.make_request(
        f"<get_notices_public><seqno>{start}</seqno></get_notices_public>"
    )
    rpc_json = xmltodict.parse(rpc_resp, force_list="notice")
    return Notices().load(rpc_json)


def get_all_notices(client: RpcClient, start: int = 1) -> dict:
    """Returns both private and non-private notices."""
    rpc_resp = client.make_request(f"<get_notices><seqno>{start}</seqno></get_notices>")
    rpc_json = xmltodict.parse(rpc_resp, force_list="notice")
    return Notices().load(rpc_json)



from marshmallow import Schema, fields, pre_load

from boinc_client.models.helpers import flatten_data, normalise_none_to_list


class PersistentFileTransfer(Schema):
    first_request_time = fields.Str()
    is_upload = fields.Str()
    last_bytes_xferred = fields.Str()
    next_request_time = fields.Str()
    num_retries = fields.Str()
    time_so_far = fields.Str()


class FileTransferDetail(Schema):
    bytes_xferred = fields.Str()
    file_offset = fields.Str()
    url = fields.Str()
    xfer_speed = fields.Str()


class FileTransfer(Schema):
    file_xfer = fields.Nested(FileTransferDetail())
    max_nbytes = fields.Str()
    name = fields.Str()
    nbytes = fields.Str()
    persistent_file_xfer = fields.Nested(PersistentFileTransfer())
    project_name = fields.Str()
    project_url = fields.Str()
    status = fields.Str()


class FileTransfers(Schema):
    file_transfers = fields.Nested(FileTransfer(many=True))

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "file_transfers")

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "file_transfers", "file_transfer")



from boinc_client.clients.rpc_client import RpcClient

from .messages import get_all_notices, message_count, messages, public_notices
from .modes import set_cpu_run_mode, set_gpu_run_mode, set_network_mode
from .preferences import (
    get_global_prefs_file,
    get_global_prefs_override,
    get_global_prefs_working,
    read_global_prefs_override,
    set_global_prefs_override,
    update_global_prefs_override,
)
from .projects import (
    all_projects,
    attach_project,
    detach_project,
    poll_attach_project,
    project_allow_more_work,
    project_no_more_work,
    reset_project,
    resume_project,
    suspend_project,
    update_project,
)
from .results import old_results, results
from .stats import daily_network_transfers, project_stats
from .status import (
    cc_status,
    client_state,
    disk_stats,
    file_transfers,
    host_info,
    project_status,
    screensaver_tasks,
    simple_gui_info,
)
from .versions import client_update, client_version


class Boinc:
    rpc_client: RpcClient

    def __init__(self, rpc_client: RpcClient):
        self.rpc_client = rpc_client

    ###########
    # Messages
    ###########
    def get_messages(self, start: int = 0) -> dict:
        return messages(client=self.rpc_client, start=start)

    def get_message_count(self) -> dict:
        return message_count(client=self.rpc_client)

    def get_public_notices(self, start: int = 0) -> dict:
        return public_notices(client=self.rpc_client, start=start)

    def get_all_notices(self, start: int = 0) -> dict:
        return get_all_notices(client=self.rpc_client, start=start)

    ###########
    # Projects
    ###########
    def get_all_projects(self) -> dict:
        return all_projects(client=self.rpc_client)

    def attach_project(self, name: str, url: str, key: str) -> dict:
        return attach_project(self.rpc_client, name, url, key)

    def poll_attach_project(self):
        return poll_attach_project(self.rpc_client)

    def update_project(self, url: str) -> dict:
        return update_project(self.rpc_client, url)

    def detach_project(self, url: str) -> dict:
        return detach_project(self.rpc_client, url)

    def suspend_project(self, url: str) -> dict:
        return suspend_project(self.rpc_client, url)

    def resume_project(self, url: str) -> dict:
        return resume_project(self.rpc_client, url)

    def reset_project(self, url: str) -> dict:
        return reset_project(self.rpc_client, url)

    ##########
    # Results
    ##########
    def get_results(self, active_only: bool = False) -> dict:
        return results(client=self.rpc_client, active_only=active_only)

    def get_old_results(self) -> dict:
        return old_results(client=self.rpc_client)

    ########
    # Stats
    ########
    def get_network_stats(self) -> dict:
        return daily_network_transfers(client=self.rpc_client)

    def get_project_stats(self) -> dict:
        return project_stats(client=self.rpc_client)

    #########
    # Status
    #########
    def get_client_state(self) -> dict:
        return client_state(client=self.rpc_client)

    def get_project_status(self) -> dict:
        return project_status(client=self.rpc_client)

    def get_cc_status(self) -> dict:
        return cc_status(client=self.rpc_client)

    def get_disk_stats(self) -> dict:
        return disk_stats(client=self.rpc_client)

    def get_file_transfers(self) -> dict:
        return file_transfers(client=self.rpc_client)

    def get_host_info(self) -> dict:
        return host_info(client=self.rpc_client)

    def get_simple_gui_info(self) -> dict:
        return simple_gui_info(client=self.rpc_client)

    def get_screensaver_tasks(self) -> dict:
        return screensaver_tasks(client=self.rpc_client)

    ########
    # Tasks
    ########
    def project_no_more_work(self, url: str) -> dict:
        return project_no_more_work(client=self.rpc_client, project_url=url)

    def project_allow_more_work(self, url: str) -> dict:
        return project_allow_more_work(client=self.rpc_client, project_url=url)

    ###########
    # Versions
    ###########
    def get_client_version(self) -> dict:
        return client_version(client=self.rpc_client)

    def get_client_update(self) -> dict:
        return client_update(client=self.rpc_client)

    ##############
    # Preferences
    ##############
    def get_global_prefs_file(self) -> dict:
        return get_global_prefs_file(self.rpc_client)

    def get_global_prefs_override(self) -> dict:
        return get_global_prefs_override(self.rpc_client)

    def set_global_prefs_override(self, pref_override: dict) -> None:
        return set_global_prefs_override(self.rpc_client, pref_override)

    def update_global_prefs_override(self, pref_override: dict) -> None:
        return update_global_prefs_override(self.rpc_client, pref_override)

    def get_global_prefs_working(self) -> dict:
        return get_global_prefs_working(self.rpc_client)

    def read_global_prefs_override(self) -> dict:
        return read_global_prefs_override(self.rpc_client)

    def set_cpu_run_mode(self, run_mode: str, duration: int = 0) -> dict:
        return set_cpu_run_mode(self.rpc_client, run_mode, duration)

    def set_gpu_run_mode(self, run_mode: str, duration: int = 0) -> dict:
        return set_gpu_run_mode(self.rpc_client, run_mode, duration)

    def set_network_mode(self, run_mode: str, duration: int = 0) -> dict:
        return set_network_mode(self.rpc_client, run_mode, duration)



import xmltodict

from boinc_client.clients.rpc_client import RpcClient
from boinc_client.models.cc_status import CCStatus
from boinc_client.models.client_state import ClientState
from boinc_client.models.disk_stats import DiskUsage
from boinc_client.models.file_transfer import FileTransfers
from boinc_client.models.host_info import HostInfo
from boinc_client.models.project_status import ProjectStatus
from boinc_client.models.screensaver_tasks import ScreensaverTasks
from boinc_client.models.simple_gui_info import SimpleGuiInfo


def client_state(client: RpcClient) -> dict:
    """Get the entire state."""
    rpc_resp = client.make_request("<get_state/>")
    rpc_json = xmltodict.parse(
        rpc_resp,
        force_list=(
            "app",
            "app_version",
            "gui_url",
            "result",
            "workunit",
        ),
    )
    return ClientState().load(rpc_json)


def project_status(client: RpcClient) -> dict:
    """Show status of all attached projects."""
    rpc_resp = client.make_request("<get_project_status/>")
    rpc_json = xmltodict.parse(rpc_resp, force_list=("project", "gui_url"))
    return ProjectStatus().load(rpc_json)


def cc_status(client: RpcClient) -> dict:
    """Show CPU/GPU/network run modes and network connection status."""
    rpc_resp = client.make_request("<get_cc_status/>")
    rpc_json = xmltodict.parse(rpc_resp)
    return CCStatus().load(rpc_json)


def disk_stats(client: RpcClient) -> dict:
    """Show disk usage by project."""
    rpc_resp = client.make_request("<get_disk_usage/>")
    rpc_json = xmltodict.parse(rpc_resp, force_list=("project",))
    return DiskUsage().load(rpc_json)


def file_transfers(client: RpcClient) -> dict:
    """Show all current file transfers."""
    rpc_resp = client.make_request("<get_file_transfers/>")
    rpc_json = xmltodict.parse(rpc_resp, force_list=("file_transfer",))
    return FileTransfers().load(rpc_json)


def host_info(client: RpcClient) -> dict:
    """Get information about host hardware and usage."""
    rpc_resp = client.make_request("<get_host_info/>")
    rpc_json = xmltodict.parse(rpc_resp)
    return HostInfo().load(rpc_json)


def simple_gui_info(client: RpcClient) -> dict:
    """Show status of projects and active tasks."""
    rpc_resp = client.make_request("<get_simple_gui_info/>")
    rpc_json = xmltodict.parse(rpc_resp, force_list=("project", "result", "gui_url"))
    return SimpleGuiInfo().load(rpc_json)


def screensaver_tasks(client: RpcClient) -> dict:
    """Show status of projects and active tasks."""
    rpc_resp = client.make_request("<get_screensaver_tasks/>")
    rpc_json = xmltodict.parse(rpc_resp, force_list=("result",))
    return ScreensaverTasks().load(rpc_json)



from marshmallow import Schema, fields


class Status(Schema):
    ams_password_error = fields.Int()
    disallow_attach = fields.Int()
    gpu_mode = fields.Int()
    gpu_mode_delay = fields.Float()
    gpu_mode_perm = fields.Int()
    gpu_suspend_reason = fields.Int()
    max_event_log_lines = fields.Int()
    network_mode = fields.Int()
    network_mode_delay = fields.Float()
    network_mode_perm = fields.Int()
    network_status = fields.Int()
    network_suspend_reason = fields.Int()
    simple_gui_only = fields.Int()
    task_mode = fields.Int()
    task_mode_delay = fields.Float()
    task_mode_perm = fields.Int()
    task_suspend_reason = fields.Int()


class CCStatus(Schema):
    cc_status = fields.Nested(Status())



from marshmallow import Schema, fields, post_load, pre_load

from boinc_client.models.helpers import (
    create_indexes,
    flatten_data,
    normalise_none_to_list,
    remove_key,
)


class Message(Schema):
    body = fields.Str(allow_none=True)
    pri = fields.Str()
    project = fields.Str(allow_none=True)
    seqno = fields.Int()
    time = fields.Int()

    @post_load
    def _a_remove_seqno(self, data, **kwargs):
        return remove_key(data, "seqno")


class Messages(Schema):
    messages = fields.Dict(fields.Int(), fields.Nested(Message()), data_key="msgs")

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "msgs")

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "msgs", "msg")

    @pre_load
    def _c_create_dict_keys(self, data, **kwargs):
        return create_indexes(data, "msgs", "seqno")



import xmltodict
from marshmallow import ValidationError

from boinc_client.clients.rpc_client import RpcClient
from boinc_client.models.generic_response import GenericResponse
from boinc_client.models.global_preference_override import GlobalPreferenceOverrides
from boinc_client.models.global_preferences import GlobalPreferences


def get_global_prefs_file(client: RpcClient) -> dict:
    """Get the contents of the global_prefs.xml file if present."""
    rpc_resp = client.make_request("<get_global_prefs_file/>")
    rpc_json = xmltodict.parse(rpc_resp)
    try:
        return GlobalPreferences().load(rpc_json)
    except ValidationError:
        return GenericResponse().load(rpc_json)


def get_global_prefs_override(client: RpcClient):
    """Get the contents of the global_prefs_override.xml file if present."""
    rpc_resp = client.make_request("<get_global_prefs_override/>")
    rpc_json = xmltodict.parse(rpc_resp)
    try:
        return GlobalPreferenceOverrides().load(rpc_json)
    except ValidationError:
        return GenericResponse().load(rpc_json)


def get_global_prefs_working(client: RpcClient):
    """Get the currently used global_prefs."""
    rpc_resp = client.make_request("<get_global_prefs_working/>")
    rpc_json = xmltodict.parse(rpc_resp)
    return GlobalPreferences().load(rpc_json)


def set_global_prefs_override(client: RpcClient, override: dict):
    """Write the global_prefs_override.xml file."""
    override_xml = "".join([f"<{k}>{v}</{k}>" for k, v in override.items()])
    rpc_resp = client.make_request(
        f"""<set_global_prefs_override>
        <global_preferences>
        {override_xml}
        </global_preferences>
        </set_global_prefs_override>"""
    )
    rpc_json = xmltodict.parse(rpc_resp)
    return GenericResponse().load(rpc_json)


def read_global_prefs_override(client: RpcClient):
    """Read the global_prefs_override.xml file and set the preferences accordingly."""
    rpc_resp = client.make_request("<read_global_prefs_override/>")
    rpc_json = xmltodict.parse(rpc_resp)
    return GenericResponse().load(rpc_json)


############
# Helpers
############
def update_global_prefs_override(client: RpcClient, override: dict):
    """Helper for updating global prefs without resetting others."""
    current_overrides = get_global_prefs_override(client)
    if "error" in current_overrides:
        merged_overrides = override
    else:
        merged_overrides = {**current_overrides["global_preferences"], **override}
    set_global_prefs_override(client, merged_overrides)



from marshmallow import Schema, fields

from boinc_client.models.result import Result


class ScreensaverTask(Schema):
    results = fields.Nested(Result(many=True), data_key="result", load_default=[])
    suspend_reason = fields.Int()


class ScreensaverTasks(Schema):
    screensaver_tasks = fields.Nested(
        ScreensaverTask(), data_key="handle_get_screensaver_tasks"
    )



from marshmallow import Schema, fields, post_load, pre_load

from boinc_client.models.helpers import (
    create_lists,
    normalise_none_to_list,
    replace_none_string,
)
from boinc_client.models.host_info import Host
from boinc_client.models.project_status import ProjectState
from boinc_client.models.result import Result


class NetStats(Schema):
    avg_down = fields.Float()
    avg_time_down = fields.Float()
    avg_time_up = fields.Float()
    avg_up = fields.Float()
    bwdown = fields.Float()
    bwup = fields.Float()


class TimeStats(Schema):
    active_frac = fields.Float()
    client_start_time = fields.Float()
    connected_frac = fields.Float()
    cpu_and_network_available_frac = fields.Float()
    gpu_active_frac = fields.Float()
    now = fields.Float()
    on_frac = fields.Float()
    previous_uptime = fields.Float()
    session_active_duration = fields.Float()
    session_gpu_active_duration = fields.Float()
    total_active_duration = fields.Float()
    total_duration = fields.Float()
    total_gpu_active_duration = fields.Float()
    total_start_time = fields.Float()


class AppDetails(Schema):
    name = fields.Str()
    non_cpu_intensive = fields.Bool()
    user_friendly_name = fields.Str()


class AppFileRef(Schema):
    copy_file = fields.Bool()
    file_name = fields.Str()
    main_program = fields.Bool()
    open_name = fields.Str(allow_none=True)

    @pre_load
    def _set_main_program(self, data, **kwargs):
        data["main_program"] = "main_program" in data
        return data

    @pre_load
    def _set_copy_file(self, data, **kwargs):
        data["copy_file"] = "copy_file" in data
        return data


class AppVersion(Schema):
    api_version = fields.Str()
    app_name = fields.Str()
    avg_ncpus = fields.Float()
    file_ref = fields.Nested(AppFileRef(many=True))
    flops = fields.Float()
    platform = fields.Str()
    version_num = fields.Int()


class WorkUnit(Schema):
    app_name = fields.Str()
    command_line = fields.Str()
    file_ref = fields.Nested(AppFileRef(many=True))
    name = fields.Str()
    rsc_disk_bound = fields.Float()
    rsc_fpops_bound = fields.Float()
    rsc_fpops_est = fields.Float()
    rsc_memory_bound = fields.Float()
    version_num = fields.Int()


class GlobalPreferences(Schema):
    battery_charge_min_pct = fields.Float()
    battery_max_temperature = fields.Float()
    confirm_before_connecting = fields.Int()
    cpu_scheduling_period_minutes = fields.Float()
    cpu_usage_limit = fields.Float()
    daily_xfer_limit_mb = fields.Float()
    daily_xfer_period_days = fields.Int()
    disk_interval = fields.Float()
    disk_max_used_gb = fields.Float()
    disk_max_used_pct = fields.Float()
    disk_min_free_gb = fields.Float()
    dont_verify_images = fields.Int()
    end_hour = fields.Float()
    hangup_if_dialed = fields.Int()
    idle_time_to_run = fields.Float()
    leave_apps_in_memory = fields.Int()
    max_bytes_sec_down = fields.Float()
    max_bytes_sec_up = fields.Float()
    max_cpus = fields.Int()
    max_ncpus_pct = fields.Float()
    mod_time = fields.Float()
    net_end_hour = fields.Float()
    net_start_hour = fields.Float()
    network_wifi_only = fields.Int()
    override_file_present = fields.Int()
    ram_max_used_busy_pct = fields.Float()
    ram_max_used_idle_pct = fields.Float()
    run_gpu_if_user_active = fields.Int()
    run_if_user_active = fields.Int()
    run_on_batteries = fields.Int()
    source_project = fields.Str(allow_none=True)
    start_hour = fields.Float()
    suspend_cpu_usage = fields.Float()
    suspend_if_no_recent_input = fields.Float()
    vm_max_used_pct = fields.Float()
    work_buf_additional_days = fields.Float()
    work_buf_min_days = fields.Float()

    @post_load
    def _a_replace_none_string(self, data, **kwargs):
        return replace_none_string(data)


class State(Schema):
    app_versions = fields.Nested(
        AppVersion(many=True), data_key="app_version", allow_none=True
    )
    apps = fields.Nested(AppDetails(many=True), data_key="app", allow_none=True)
    core_client_major_version = fields.Int()
    core_client_minor_version = fields.Int()
    core_client_release = fields.Int()
    executing_as_daemon = fields.Bool()
    global_preferences = fields.Nested(GlobalPreferences())
    host_info = fields.Nested(Host())
    net_stats = fields.Nested(NetStats())
    platform = fields.Str()
    platform_name = fields.Str()
    projects = fields.Nested(
        ProjectState(many=True), data_key="project", load_default=[]
    )
    results = fields.Nested(Result(many=True), data_key="result", load_default=[])
    time_stats = fields.Nested(TimeStats())
    work_units = fields.Nested(
        WorkUnit(many=True), data_key="workunit", allow_none=True
    )

    @pre_load
    def _a_create_keys(self, data, **kwargs):
        return create_lists(data, ["app", "app_version", "project", "workunit"])

    @pre_load
    def _b_wrap_dict(self, data, **kwargs):
        data["project"] = (
            [data["project"]] if type(data["project"]) is not list else data["project"]
        )
        return data


class ClientState(Schema):
    client_state = fields.Nested(State())



from marshmallow import Schema, fields, post_load, pre_load

from boinc_client.models.helpers import (
    create_lists,
    flatten_data,
    normalise_none_to_list,
    replace_none_string,
    set_bools,
)


class RscBackoffTime(Schema):
    name = fields.Str()
    value = fields.Float()


class RscBackoffInterval(Schema):
    name = fields.Str()
    value = fields.Float()


class GuiUrl(Schema):
    description = fields.Str()
    name = fields.Str()
    url = fields.Url()


class ProjectState(Schema):
    cpid_time = fields.Float()
    cross_project_id = fields.Str(allow_none=True)
    desired_disk_usage = fields.Float()
    disk_share = fields.Float()
    disk_usage = fields.Float()
    dont_request_more_work = fields.Bool()
    dont_use_dcf = fields.Bool()
    duration_correction_factor = fields.Float()
    elapsed_time = fields.Float()
    email_hash = fields.Str(allow_none=True)
    external_cpid = fields.Str(allow_none=True)
    gui_urls = fields.Nested(GuiUrl(many=True))
    host_create_time = fields.Float()
    host_expavg_credit = fields.Float()
    host_total_credit = fields.Float()
    host_venue = fields.Str(allow_none=True)
    hostid = fields.Int()
    last_rpc_time = fields.Float()
    master_fetch_failures = fields.Int()
    master_url = fields.Str()
    master_url_fetch_pending = fields.Bool()
    min_rpc_time = fields.Float()
    next_rpc_time = fields.Float()
    njobs_error = fields.Int()
    njobs_success = fields.Int()
    nrpc_failures = fields.Int()
    project_dir = fields.Str()
    project_files_downloaded_time = fields.Float()
    project_name = fields.Str()
    rec = fields.Float()
    rec_time = fields.Float()
    resource_share = fields.Float()
    rpc_seqno = fields.Int()
    rsc_backoff_interval = fields.Nested(RscBackoffInterval())
    rsc_backoff_time = fields.Nested(RscBackoffTime())
    sched_priority = fields.Float()
    sched_rpc_pending = fields.Int()
    scheduler_rpc_in_progress = fields.Bool()
    send_job_log = fields.Int()
    send_time_stats_log = fields.Int()
    suspended_via_gui = fields.Bool()
    symstore = fields.Str(allow_none=True)
    team_name = fields.Str(allow_none=True)
    teamid = fields.Int()
    upload_backoff = fields.Float()
    user_create_time = fields.Float()
    user_expavg_credit = fields.Float()
    user_name = fields.Str(allow_none=True)
    user_total_credit = fields.Float()
    userid = fields.Int()

    @pre_load
    def _a_create_keys(self, data, **kwargs):
        return create_lists(data, ["gui_urls"])

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "gui_urls", "gui_url")

    @pre_load
    def _c_set_bools(self, data, **kwargs):
        return set_bools(
            data,
            [
                "dont_request_more_work",
                "dont_use_dcf",
                "master_url_fetch_pending",
                "scheduler_rpc_in_progress",
                "suspended_via_gui",
            ],
        )

    @post_load
    def _a_replace_none_string(self, data, **kwargs):
        return replace_none_string(data)


class ProjectStatus(Schema):
    project_status = fields.Nested(ProjectState(many=True), data_key="projects")

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "projects")

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "projects", "project")



from marshmallow import Schema, fields


class GlobalPreference(Schema):
    battery_charge_min_pct = fields.Float()
    battery_max_temperature = fields.Float()
    confirm_before_connecting = fields.Int()
    cpu_scheduling_period_minutes = fields.Float()
    cpu_usage_limit = fields.Float()
    daily_xfer_limit_mb = fields.Float()
    daily_xfer_period_days = fields.Int()
    disk_interval = fields.Float()
    disk_max_used_gb = fields.Float()
    disk_max_used_pct = fields.Float()
    disk_min_free_gb = fields.Float()
    dont_verify_images = fields.Int()
    end_hour = fields.Float()
    hangup_if_dialed = fields.Int()
    idle_time_to_run = fields.Float()
    leave_apps_in_memory = fields.Int()
    max_bytes_sec_down = fields.Float()
    max_bytes_sec_up = fields.Float()
    max_cpus = fields.Int()
    max_ncpus_pct = fields.Float()
    mod_time = fields.Float()
    net_end_hour = fields.Float()
    net_start_hour = fields.Float()
    network_wifi_only = fields.Int()
    override_file_present = fields.Int()
    ram_max_used_busy_pct = fields.Float()
    ram_max_used_idle_pct = fields.Float()
    run_gpu_if_user_active = fields.Int()
    run_if_user_active = fields.Int()
    run_on_batteries = fields.Int()
    source_project = fields.Str(allow_none=True)
    start_hour = fields.Float()
    suspend_cpu_usage = fields.Float()
    suspend_if_no_recent_input = fields.Float()
    vm_max_used_pct = fields.Float()
    work_buf_additional_days = fields.Float()
    work_buf_min_days = fields.Float()


class GlobalPreferences(Schema):
    global_preferences = fields.Nested(GlobalPreference())



import xmltodict

from boinc_client.clients.rpc_client import RpcClient
from boinc_client.models.old_results import OldResults
from boinc_client.models.result import Results


def results(client: RpcClient, active_only: bool = False) -> dict:
    """Show tasks."""
    rpc_resp = client.make_request(
        f"<get_results><active_only>{int(active_only)}</active_only></get_results>"
    )
    rpc_json = xmltodict.parse(rpc_resp, force_list="result")
    return Results().load(rpc_json)


def old_results(client: RpcClient) -> dict:
    """Show old tasks."""
    rpc_resp = client.make_request("<get_old_results/>")
    rpc_json = xmltodict.parse(rpc_resp, force_list="old_result")
    return OldResults().load(rpc_json)
